var ElementJSON = (function () {
    'use strict';
    /**
     * @description This is the main object, he is used to manipulate the elements.
     * @param {object[]} json_object_tableau Array of object compatible with cytoscape.js.
     * @typedef ElementJSON Contains all of the elements for cytoscape.js, manipulate the elements in order to provide good data for the graph visualization.
     */
    function ElementJSON(json_object_tableau) {
        /**
         * @description Contains all the json for the graph.
         * @type {{origin:object, groupbycause:object, groupbyimpact:object, filter:object, toArray:function}}
         */
        var json = {
            origin: undefined,
            groupbycause: undefined,
            groupbyimpact: undefined,
            filter: null,
            toArray: function (jsonToUse) {
                return $.map(jsonToUse.nodes, function (element) {
                    return element.basicElement();
                }).concat($.map(jsonToUse.groups, function (element) {
                    return element.basicElement();
                }).concat($.map(jsonToUse.edges, function (element) {
                    return element.basicElement();
                })));
            }
        };
        /**
         * @description Store the current json used.
         * @type string
         */
        var used = utilisation.origin;
        /**
         * @description Contains the root nodes.
         * @type {GNode[]}
         */
        var rootsNode = [];
        /**
         * @description Contains the end nodes.
         * @type {GNode[]}
         */
        var endNode = [];
        /**
         * @description Contains all the selected nodes.
         * @type {GNode[]}
         */
        var selectedNode = [];
        /**
         * @description We used viz.js to get the position of the node because the viz' layout is more smart than cytoscape's layout.
         * @type VizElement
         */
        var viz_element;

        /**
         * @description Get the json used (no heed of filter usage).
         * @returns {string}
         */
        this.getUsed = function () { return used; };
        /**
         * @description Get the 'rootsNode' (the nodes that doesn't have predecessor).
         * @returns {object[]}
         */
        this.getRoots = function () { return rootsNode; };
        /**
         * @description Get the json 'origin' element.
         * @returns {object[]}
         */
        this.getElementOrigin = function () { return getElementBy(utilisation.origin); };
        /**
         * @description Get the json 'groupbycause' element (element grouped by cause).
         * @returns {object[]}
         */
        this.getElementByCause = function () { return getElementBy(utilisation.cause); };
        /**
         * @description Get the json 'groupbyimpact' element (element grouped by impact).
         * @returns {object[]}
         */
        this.getElementByImpact = function () { return getElementBy(utilisation.impact); };
        /**
         * @description Apply a filter on the json used and get him.
         * @param {string[]} valArray Array of the node's class to filter.
         * @returns {object[]}
         */
        this.getFilteredElement = function (valArray) {
            if (valArray !== undefined) {
                json.filter = {nodes: shallowCopy(json[used].nodes), edges: shallowCopy(json[used].edges), groups: shallowCopy(json[used].groups)};
                filter(valArray);
            }
            return json.toArray(json.filter);
        };
        /**
         * @description Give the json used (no heed of filter json).
         * @returns {object[]}
         */
        this.getUsedJson = function () { return getElementBy(); };
        /**
         * @description Give the json currently used (heed of the filter json).
         * @returns {object[]}
         */
        this.getElementToUse = function () { return json.toArray(getJsonToUse()); };
        /**
         * @description Give the element for the usage wanted (no heed of filter usage.).
         * @param {string} usage_wanted
         * @returns {object[]}
         */
        this.getElementFor = function (usage_wanted) { return getElementBy(usage_wanted); };
        /**
         * @description Give the number of node for the json to use (heed of the filter json).
         * @returns {number}
         */
        this.getNumberOfNodesForJsonToUse = function () { return Object.keys(getJsonToUse().nodes).length; };
        /**
         * @description Give the json designated by the 'group' (no heed of filter json).
         * @param {string} [group] Default to 'used'.
         * @returns {object[]} Return null if wrong 'group'
         */
        function getElementBy (group) {
            group = group || used;
            switch (group) {
                case utilisation.origin:
                    /* If the usage is different from 'origin', we delete the data parent of each node. */
                    if (used !== utilisation.origin)
                        new IteratorGraph(json.origin.nodes).forEach(function (node) {
                            delete node.data()[node_a.data.parent];
                        }, function (percentage) {console.log("Deleting nodes parents: " + percentage.toString().substring(0,4) + "% clear.");});
                    break;
                case utilisation.cause:
                case utilisation.impact:
                    /* We create the grouped json once. If it already exist, we just update the parent of each node. */
                    if(!json[group]) createJsonGroupBy(group);
                    else                  updateParent(group);
                    break;
                default:
                    return null;
            }
            /* If there was a filter json, we turned him null. */
            if (json.filter) {
                json.filter = null;
                new IteratorGraph(json.origin.nodes).forEach(function (node) { node.removeProp(node_p.filter); });
            }
            /* Update the usage and return the json. */
            used = group;
            return json.toArray(json[used]);
        }
        /**
         * @description Give the json currently used (heed of the filter json).
         * @returns {{nodes:object, edges:object[]: groups:object[]}|null}
         */
        function getJsonToUse() {
            if (!json.filter) {
                switch (used) {
                    case utilisation.origin:
                    case utilisation.cause:
                    case utilisation.impact:
                        return json[used];
                    default:
                        return null;
                }
            } else {
                return json.filter;
            }
        }
        /**
         * @description Give the settings of the current instance of VizElement.
         *      - For now we just need the engine.
         * @returns {string}
         */
        this.getVizLayout = function () { return viz_element.getEngine(); };

        /**
         * @description Tell if the json is filtered.
         * @returns {boolean}
         */
        this.isFilterJson = function () { return json.filter !== null; };

        /**
         * @description Complete initialization of the original json (the 'origin' json).
         */
        function initJsonElement() {
            var elements_to_chain = preparing_element_to_chain();
            chaining_the_elements(elements_to_chain);
            initialize_json_origin();
            initViz();
        }
        /**
         * @description Preparing the element to their future chaining.
         * @returns {object}
         */
        function preparing_element_to_chain() {
            var element_to_chain = {};

            new IteratorGraph(json_object_tableau).forEach(function (element) {
                if (ele.is(element, elementGroup.node)) {
                    element = new GNode(element);

                    /* Adding the node in 'element_to_chain'. */
                    if (element_to_chain[element.id()])    element_to_chain[element.id()].himself = element;
                    else                                   element_to_chain[element.id()] = {himself: element};

                    /* Adding the node's class to the proper property. */
                    element.setAttr(node_a.classes, element.data(node_a.data.class).toLowerCase());
                    element.addClass(css_classes.external);
                } else if (ele.is(element, elementGroup.edge)) {
                    element = new GEdge(element);
                    /* Storing the edge source and target (corresponding to node id). */
                    var id_node_source = element.data(edge_a.data.source), id_node_target = element.data(edge_a.data.target);

                    /* Adding the edge as successor and target of the appropriate node in 'element_to_chain'. */
                    if (element_to_chain[id_node_source]) {
                        if (element_to_chain[id_node_source].edges)
                            element_to_chain[id_node_source].edges.succeed.push(element);
                        else
                            element_to_chain[id_node_source].edges = {succeed: [element], precede: []};
                    } else {
                        element_to_chain[id_node_source] = {edges: {succeed: [element], precede: []}};
                    }
                    if (element_to_chain[id_node_target]) {
                        if (element_to_chain[id_node_target].edges) {
                            element_to_chain[id_node_target].edges.precede.push(element);
                        } else {
                            element_to_chain[id_node_target].edges = {succeed: [], precede: [element]};
                        }
                    } else {
                        element_to_chain[id_node_target] = {edges: {succeed: [], precede: [element]}};
                    }
                }
            }, function (percentage) { console.log("Initialization first step: " + percentage.toString().substring(0,4) + "% clear."); });

            return element_to_chain;
        }
        /**
         * @description Chaining the given elements.
         * @param {object} elements_to_chain
         */
        function chaining_the_elements(elements_to_chain) {
            var iterator = new IteratorGraph(elements_to_chain);

            iterator.forEach(function (element) {
                var node = element.himself;

                if (element.edges) {
                    /* Linking the preceding edges if they are. */
                    if (element.edges.precede.length > 0) {
                        node.setProp(node_p.precedingEdges, element.edges.precede);
                        /* Adding the node as target of the edges. */
                        for (var precedingEdgeArr = element.edges.precede, pe = 0, precedingEdge = precedingEdgeArr[pe];
                             pe < precedingEdgeArr.length;
                             pe++, precedingEdge = precedingEdgeArr[pe])
                        {
                            precedingEdge.setProp(edge_p.target, node);
                        }
                    } else {
                        /* A node without preceding edge is a root node. */
                        rootsNode.push(node);
                    }

                    /* Linking the succeeding edges if they are. */
                    if (element.edges.succeed.length > 0) {
                        node.setProp(node_p.succeedingEdges, element.edges.succeed);
                        /* Adding the node as source of the edges. */
                        for (var succeedingEdgeArr = element.edges.succeed, se = 0, succeedingEdge = succeedingEdgeArr[se];
                             se < succeedingEdgeArr.length;
                             se++, succeedingEdge = succeedingEdgeArr[se])
                        {
                            succeedingEdge.setProp(edge_p.source, node);
                        }
                    } else {
                        /* A node without succeeding edge is a end node. */
                        endNode.push(node);
                    }
                } else {
                    /* A node without edges is the root and end node of his impact. */
                    rootsNode.push(node); endNode.push(node);
                }
            }, function (percentage) {
                console.log("Initialization second step: " + percentage.toString().substring(0,4) + "% clear.");
            });
        }
        /**
         * @description Initializing the json origin with the previously prepared data. Starting by the root nodes.
         */
        function initialize_json_origin() {
            json.origin = {nodes: {}, edges: [], groups: []};
            fill_json_origin(rootsNode);
        }
        /**
         * @description Fill the json origin with all the known data.
         * @param {object[]} nodes_to_add
         */
        function fill_json_origin(nodes_to_add) {
            var succeeding_nodes = [];

            if (nodes_to_add.length)    succeeding_nodes = fill_json_from_array(nodes_to_add);

            if (succeeding_nodes.length > 0) fill_json_origin(succeeding_nodes);
        }
        /**
         * @description Fill the json from each node contained in the 'node_array'.
         * @param {object[]} node_array
         * @returns {object[]} Return an empty array if node succeeding nodes.
         */
        function fill_json_from_array(node_array) {
            var succeeding_nodes = [];
            for (var n = 0; n < node_array.length; n++)
                succeeding_nodes = succeeding_nodes.concat(fill_json_from_object(node_array[n]));
            return succeeding_nodes;
        }
        /**
         * @description Fill the json with this node.
         * @param {object} node
         * @returns {object[]}
         */
        function fill_json_from_object(node) {
            /* Adding this node in origin. */
            if (!json.origin.nodes[node.id()]) {
                json.origin.nodes[node.id()] = node;

                /* Verify if this node has successors, add the succeeding edges to origin and call return the succeeding nodes. */
                if (node.hasSuccessors()) {
                    json.origin.edges = json.origin.edges.concat(node.succeedingEdges());
                    return node.succeedingNodes();
                }
            }

            return [];
        }

        /**
         * @description Initialize 'viz_element'.
         */
        function initViz() {
            /**
             * @description Give the number of 'inch' that is equal to the number of pixel given.
             * @param {number} pixel
             * @returns {number}
             */
            function pixel2inch(pixel) {
                /* 1 pixel = 0.010416667 inch */
                return pixel * 0.010416667;
            }

            viz_element = new VizElement({
                width: pixel2inch(window.innerWidth || document.documentElement.clientWidth || document.getElementsByTagName('body')[0].clientWidth),
                height: pixel2inch(window.innerHeight || document.documentElement.clientHeight || document.getElementsByTagName('body')[0].clientHeight),
                nodeShape: "box"
            });
        }
        /**
         * @description Create the dot from the json to use for 'viz_element' (heed of the filter json).
         * @returns {string}
         */
        function createDotForViz(jsonToUse) {
            var response = "";

            var iterator = new IteratorGraph(jsonToUse.nodes);
            iterator.forEach(function (node) {
                if (node.data(node_a.data.parent))  response += "subgraph \"cluster_" + node.data(node_a.data.parent) + "\" {\"" + node.id() + "\";}\n";
                else                                response += "\"" + node.id() + "\";\n";
            }, function (percentage) {
                console.log("Creating dot nodes for viz element: " + percentage.toString().substring(0,4) + "% clear.");
            });
            iterator.changeElement(jsonToUse.edges);
            iterator.forEach(function (edge) {
                response += "\"" + edge.data(edge_a.data.source) + "\" -> \"" + edge.data(edge_a.data.target) + "\";\n";
            }, function (percentage) {
                console.log("Creating dot edges for viz element: " + percentage.toString().substring(0,4) + "% clear.");
            });

            return response;
        }
        /**
         * @description Update the nodes position from 'viz_element'.
         * @see VizElement
         */
        function majNodePositionFromViz() {
            /**
             * @description Give the number of 'pixel' that is equel to the number of inch given.
             * @param {number} inch
             * @returns {number}
             */
            function inch2pixel(inch) {
                /* 1 inch = 96 pixel */
                return inch * 96;
            }
            var width = window.innerWidth || document.documentElement.clientWidth || document.getElementsByTagName('body')[0].clientWidth,
                height = window.innerHeight || document.documentElement.clientHeight || document.getElementsByTagName('body')[0].clientHeight;
            var position = viz_element.getPosition();

            var node_iterator = new IteratorGraph(getJsonToUse().nodes);
            node_iterator.forEach(function (node) {
                var nodeId = node.id();
                if (position.nodes[nodeId]) {
                    node.setProp(node_p.width, position.nodes[nodeId].width);
                    node.setProp(node_p.height, position.nodes[nodeId].height);
                    node.setAttr(node_a.position, {
                        x: (width * inch2pixel(position.nodes[nodeId].x) / inch2pixel(position.graph.width) ),
                        y: (height * inch2pixel(position.nodes[nodeId].y) / inch2pixel(position.graph.height) )
                    });
                }
            }, function (percentage) {
                console.log("Updating nodes position from viz: " + percentage.toString().substring(0,4) + "% clear.");
            });
        }
        /**
         * @description Change the Viz engine.
         * @param {string} engine
         * @see VizElement
         */
        this.changeVizEngine = function (engine) { viz_element.setEngine(engine); };
        /**.
         * @description Execute Viz.js with the current settings and update the nodes position.
         */
        this.runViz = function () {
            var elementToReturn = {ok : true, element : []};
            var jsonToUse = getJsonToUse();

            if (Object.keys(jsonToUse.nodes).length > sizeToConfirm) elementToReturn.ok = popup_confirm(Object.keys(jsonToUse.nodes).length);

            if (elementToReturn.ok) {
                /*Donne à VizElement son graphe sous forme de string.*/
                /* Create and give the dot for Viz.js and run it with the current settings. */
                viz_element.setElement(createDotForViz(jsonToUse));
                viz_element.run();
                /* Update the nodes position from Viz.js. */
                majNodePositionFromViz();
                /* Give the result as an array. */
                elementToReturn.element = json.toArray(getJsonToUse());
            }
            return elementToReturn;
        };

        /**
         * @description Copy the json. Copy this 'json' key/value and take reference of the values if they are not primitive type.
         * @param {object|Array} json
         * @return {object|Array}
         */
        function shallowCopy(json) {
            var copy = (json.length ? [] : {});
            $.each(json, function (index, item) { copy[index] = item; });
            return copy;
        }
        /**
         * @description Deep copy of the JSON. Copy all of the primitive and 'function' values of this 'element', if 'object' type find goes deeper to copy the object.
         * @param {Array|object|null} element
         * @returns {Array|object|null}
         */
        function copyOf(element) {
            if (element !== null) {
                var copy = (!element.length ? {} : []);

                $.each(element, function (index, item) {
                    if (typeof item !== "object")  copy[index] = item;
                    else                           copy[index] = copyOf(item);
                });

                return copy;
            } else {
                return element;
            }
        }

        /**
         * @description Create a grouped json.
         * @param {string} group The wanted group.
         */
        function createJsonGroupBy(group) {
            json[group] = {nodes: shallowCopy(json.origin.nodes), edges: shallowCopy(json.origin.edges), groups: []};

            var groupNamesArr = [];
            var groupContentArr = {};

            var node_iterator = new IteratorGraph(json[group].nodes);
            node_iterator.forEach(function (node) {
                if (node.data(group)) {
                    var group_name = node.data(group);
                    if (!groupContentArr[group_name]) groupContentArr[group_name] = [];
                    if (groupContentArr[group_name].indexOf(node) === -1) groupContentArr[group_name].push(node);
                    if (groupNamesArr.indexOf(group_name) === -1) groupNamesArr.push(group_name);
                }
            }, function (percentage) { console.log("Create group by " + group + ": " + percentage.toString().substring(0,4) + "% clear."); });

            for (var gn = 0, groupName = groupNamesArr[gn], nodesToAdd = groupContentArr[groupName] ;
                                        gn < groupNamesArr.length ;
                 gn++, groupName = groupNamesArr[gn], nodesToAdd = groupContentArr[groupName]) {

                /*
                 * We create the group only if there are 2 or more node in the group.
                 * The question was: show a single node in a group can be an information for the user on what
                 *      the impact of this node could be or not green even if the node have an 'impact' of 100 and his edge a criticality of 100.
                 *      So do we show it or not?
                 * The answer is: it's a good idea but we fear that, with a large sub-graph, it can blurred the lecture and be uncomfortable for the user.
                 */
                if (nodesToAdd.length > 1) {
                    /*
                     * We replace "-" by "_" for compatibility with Viz.js.
                     * Indeed, Viz.js understand the "-" as a link between two nodes and there is no key word to differenciate a complete word of a node declaration.
                     * So the text "subgraph cluster_example-node" will cause an error.
                     */
                    createGroupBy(groupName.replace(/-/g, "_"), groupName, nodesToAdd, group);
                }
            }
        }
        /**
         * @description Create a group and tell all of his sons that he is their parent.
         * @param {string} groupId The group id (will be the 'parent' value).
         * @param {string} groupLabel The label of the group.
         * @param {object[]} nodes The nodes to group.
         * @param {string} group The group wanted.
         */
        function createGroupBy(groupId, groupLabel, nodes, group) {
            var nodeParent = {
                data: {
                    id: groupId,
                    label: groupLabel
                },
                group: elementGroup.node,
                classes: "expand_group"
            };
            nodeParent = new GNode(nodeParent);
            nodeParent.setProp(node_p.nodes, nodes);

            for (var n = 0, node = nodes[n] ; n < nodes.length ; n++, node = nodes[n]) {
                node.data(node_a.data.parent, groupId);
                if (!node.hasProp(node_p.parent))   node.setProp(node_p.parent, {})[group] = groupId;
                else                                node.getProp(node_p.parent)[group] = groupId;
            }

            /* We only have to add the group in the correct json. Cytoscape.js will identify the node that compose him with the data 'parent' on each node. */
            json[group].groups.push(nodeParent);
        }
        /**
         * @description Update the data 'parent' of each node if the group has already been create.
         * @param {string} group The group wanted.
         */
        function updateParent(group) {
            var node_iterator = new IteratorGraph(json.origin.nodes);
            node_iterator.forEach(function (node) {
                if (node.hasProp(node_p.parent) && node.getProp(node_p.parent)[group]) node.data(node_a.data.parent, node.getProp(node_p.parent)[group]);
            }, function (percentage) { console.log("Updating parents for " + group + ": " + percentage.toString().substring(0,4) + "% clear."); });
        }

        /**
         * @description Apply a filter on the json to hide the node of the classes specified in 'valArray'.
         * @param {string[]} valArray
         */
        function filter(valArray) {
            var nodeToBeFiltered = [];

            var node_iterator = new IteratorGraph(json.origin.nodes);
            node_iterator.forEach(function (node) {
                node.removeProp(node_p.filter);
                if (valArray.indexOf(node.data(node_a.data.class)) !== -1) nodeToBeFiltered.push(node);
            }, function (percentage) { console.log("Taken all node to be filtered: " + percentage.toString().substring(0,4) + "% clear."); });

            filterNode(nodeToBeFiltered);

            node_iterator.changeElement(json.filter.nodes);
            /* If there is groups we must verify if they currently have nodes, if not we must remove them. */
            if (json.filter.groups.length > 0) {
                var groupOk = {};

                node_iterator.forEach(function (node) {
                    if (node.data(node_a.data.parent)) {
                        if (!groupOk[node.data(node_a.data.parent)])    groupOk[node.data(node_a.data.parent)] = 1;
                        else                                            groupOk[node.data(node_a.data.parent)]++;
                    }
                }, function (percentage) { console.log("Verifying filter group first step: " + percentage.toString().substring(0,4) + "% clear."); });

                if (Object.keys(groupOk).length > 0) {
                    for (var g = json.filter.groups.length - 1, group = json.filter.groups[g];
                                                    g >= 0 ;
                                    g--, group = json.filter.groups[g])
                    {
                        if (!groupOk[group.id()]) json.filter.groups.splice(g, 1);
                    }
                } else
                    json.filter.groups = [];
            }
        }
        /**
         * @description Replace all the nodes to be filtered by edges that contains the node in 'data.nodes'.
         * @param {object[]} nodeToBeFiltered
         */
        function filterNode(nodeToBeFiltered) {
            var iterator = new IteratorGraph(nodeToBeFiltered);
            iterator.forEach(function (node) {
                /* If this node is selected, we can't filter it because the user could want to just see him and not his similar. */
                if (!node.selected()) {
                    var _ele = new GElement(node, json.filter.edges),
                        precedingEdges  = _ele.precedingEdges(),
                        succeedingEdges = _ele.succeedingEdges();

                    if (precedingEdges.length > 0 && succeedingEdges.length > 0)
                        replacingNode(node, precedingEdges, succeedingEdges);

                    /* Deleting the nodes in json.filter. */
                    delete json.filter.nodes[node.id()];
                    node.setProp(node_p.filter, true);
                }
            }, function (percentage) { console.log("Filter node: " + percentage.toString().substring(0,4) + "% clear."); });

            for (var e = json.filter.edges.length - 1, edge = json.filter.edges[e] ; e >= 0 ; e--, edge = json.filter.edges[e]) {
                if (edge.source().areYou(node_p.filter) || edge.target().areYou(node_p.filter)) json.filter.edges.splice(e, 1);
            }
        }
        /**
         * @description Replacing the node by edges.
         * @param {object} node The node to replace.
         * @param {object[]} precedingEdges Les edges qui le précede.
         * @param {object[]} succeedingEdges Les edges qui le suive.
         */
        function replacingNode (node, precedingEdges, succeedingEdges) {
            /* Foreach preceding/succeeding edge we create a edge who contains this node in 'data.nodes'. */
            for (var ep = 0, precedingEdge = precedingEdges[ep];
                                            ep < precedingEdges.length;
                 ep++, precedingEdge = precedingEdges[ep])
            {
                var precedingNode  = json.origin.nodes[precedingEdge.data(edge_a.data.source)];
                for (var es = 0, succeedingEdge = succeedingEdges[es];
                                            es < succeedingEdges.length;
                     es++, succeedingEdge = succeedingEdges[es])
                {
                    var succeedingNode = json.origin.nodes[succeedingEdge.data(edge_a.data.target)];
                    /* We create the edge only if none of the preceding/succeeding node is filter. */
                    if (!precedingNode.areYou(node_p.filter) && !succeedingNode.areYou(node_p.filter)) {
                        var replacing_edge = {
                            data: {
                                source: precedingEdge.data(edge_a.data.source),
                                target: succeedingEdge.data(edge_a.data.target)
                            }, group: elementGroup.edge, classes: "filter"
                        };
                        replacing_edge = new GEdge(replacing_edge);
                        replacing_edge.setProp(edge_p.source, precedingEdge.getProp(edge_p.source));
                        replacing_edge.setProp(edge_p.target, succeedingEdge.getProp(edge_p.target));
                        replacing_edge.setProp(edge_p.nodes, [node]);

                        /* If the preceding or/and succeeding edge already contains nodes, we must keep this information in the newly created edge. */
                        var nodes = replacing_edge.getProp(edge_p.nodes);
                        if (precedingEdge.hasProp(edge_p.nodes))   nodes = nodes.concat(precedingEdge.getProp(edge_p.nodes));
                        if (succeedingEdge.hasProp(edge_p.nodes))  nodes = nodes.concat(succeedingEdge.getProp(edge_p.nodes));

                        json.filter.edges.push(replacing_edge);
                    }

                }
            }
        }

        /**
         * @description Update the nodes data (impact, classes, etc...).
         * @param {GNode} node The node to update.
         * @param {boolean} [selection] If the update comes from a select event or a unselect event. Default to false.
         */
        this.majNode = function (node, selection) {
            selection = selection || false;

            if (selection) {
                if (selectedNode.indexOf(node) === -1) selectedNode.push(node);
                node.select();
                propagateProp(node_p.impacted, node.setProp(node_p.impacted, true), "successor", node);
                if (node.hasPredecessors()) propagateCause(node.precedingNodes(), node);

                majImpact(endNode);
                majClasses();
            } else {
                selectedNode.splice(selectedNode.indexOf(node), 1);
                node.unselect();
                if (node.hasSuccessors())   undoImpact(node.succeedingNodes());
                if (node.hasPredecessors()) undoCause(node.precedingNodes(), node);

                if (selectedNode.length > 0)    majNodes();
                else                            majClasses();
            }
        };
        /**
         * @description Update the nodes from the 'selectedNode'.
         */
        function majNodes () {
            for (var n = 0 , node = selectedNode[n] ; n < selectedNode.length ; n++, node = selectedNode[n]) {
                propagateProp(node_p.impacted, node.setProp(node_p.impacted, true), "successor", node);
                if (node.hasPredecessors()) propagateCause(node.precedingNodes(), node);
            }

            majImpact(endNode);
            majClasses();
        }
        /**
         * @description Propagate a property following a guidance.
         * @param {string} propName The name of the property to propagate.
         * @param {*} propValue The value to propagate.
         * @param {string} guidance The guidance ("successor" | "predecessor").
         * @param {GNode} node The node from which we want to propagate the property.
         */
        function propagateProp (propName, propValue, guidance, node) {
            switch (guidance) {
                case "successor":
                    if (node.hasSuccessors()) toSuccessors(propName, propValue, node.succeedingNodes());
                    break;
                case "predecessor":
                    if (node.hasPredecessors()) toPredecessors(propName, propValue, node.precedingNodes());
                    break;
                default:
                    throw "ElementJSON-propagate: Unknown guidance.";
            }
        }

        /**
         * @description Propagate a property only to the successors and only if they don't have the property.
         * @param {string} propName The name of the property to propagate.
         * @param {*} propValue The value to propagate.
         * @param {GNode|GNode[]} nodes The nodes to propagate the property.
         */
        function toSuccessors (propName, propValue, nodes) {
            if (nodes.length) {
                for (var n = 0, node = nodes[n]; n < nodes.length; n++, node = nodes[n]) {
                    node.setProp(propName, propValue);
                    if (node.hasSuccessors()) {
                        toSuccessors(propName, propValue, node.succeedingNodes());
                    }
                }
            } else {
                nodes.setProp(propName, propValue);
                if (nodes.hasSuccessors()) {
                    toSuccessors(propName, propValue, nodes.succeedingNodes());
                }
            }
        }
        /**
         * @description Propagate a property only to the predecessors.
         * @param {string} propName The name of the property to propagate.
         * @param {*} propValue The value to propagate.
         * @param {GNode|GNode[]} nodes The nodes to propagate the property.
         */
        function toPredecessors (propName, propValue, nodes) {
            if (nodes.length) {
                for (var n = 0, node = nodes[n]; n < nodes.length; n++, node = nodes[n]) {
                    node.setProp(propName, propValue);
                    if (node.hasPredecessors()) {
                        toSuccessors(propName, propValue, node.precedingNodes());
                    }
                }
            } else {
                nodes.setProp(propName, propValue);
                if (nodes.hasPredecessors()) {
                    toSuccessors(propName, propValue, nodes.precedingNodes());
                }
            }
        }
        /**
         * @description Remove the impact of this node and this succeeding nodes' impacts only if they are not selected.
         * @param {GNode|GNode[]} nodes
         */
        function undoImpact (nodes) {
            /**
             * @description Delete the impact of this node only if this node is not selected and undo this node's succeeding nodes.
             * @param {GNode} node
             */
            function undo_impact (node) {
                if (node.areYou(node_p.impacted)) {
                    node.removeProp(node_p.impacted);
                    delete node.data()[node_a.data.impactNode];

                    if (node.hasSuccessors()) undoImpact(node.succeedingNodes());
                }
            }

            if (nodes.length) {
                for (var n = 0, node = nodes[n]; n < nodes.length; n++, node = nodes[n]) undo_impact(node);
            } else
                undo_impact(nodes);
        }
        /**
         * @description Remove this node's cause and this preceding nodes' cause only if they are not impacted.
         * @param {GNode|GNode[]} nodes
         * @param {GNode} restored_node
         */
        function undoCause (nodes, restored_node) {
            /**
             * @description Decrease the cause of this node only if he is not impacted. If this node's cause is 1 then remove his cause.
             * @param {GNode} node
             */
            function undo_cause (node) {
                if (!node.areYou(node_p.impacted)) {
                    if (node.areYou(node_p.causeOf)) {
                        var failure = node.getProp(node_p.causeOf);
                        if (failure.indexOf(restored_node) !== -1) failure.splice(failure.indexOf(restored_node), 1);
                        if (failure.length === 0) node.removeProp(node_p.causeOf);

                        if (node.hasPredecessors()) undoCause(node.precedingNodes(), restored_node);
                    }
                } else
                    if (node.hasPredecessors()) undoCause(node.precedingNodes(), restored_node);
            }

            if (nodes.length) {
                for (var n = 0, node = nodes[n]; n < nodes.length; n++, node = nodes[n]) undo_cause (node);
            } else
                undo_cause(nodes);
        }
        /**
         * @description Propagate the cause to the nodes.
         * @param {GNode|GNode[]} nodes
         * @param {GNode} crushing_node
         */
        function propagateCause (nodes, crushing_node) {
            /**
             * @description Propagate the cause to this node only if this node is not impacted and propagate to this preceding nodes.
             * @param {GNode} node
             */
            function propagate_cause (node) {
                if (!node.areYou(node_p.impacted)) {
                    if (!node.hasProp(node_p.causeOf)) {
                        node.setProp(node_p.causeOf, [crushing_node]);
                        if (node.hasPredecessors()) propagateCause(node.precedingNodes(), crushing_node);
                    } else {
                        var failure = node.getProp(node_p.causeOf);
                        if (failure.indexOf(crushing_node) === -1) {
                            failure.push(crushing_node);
                            if (node.hasPredecessors()) propagateCause(node.precedingNodes(), crushing_node);
                        }
                    }
                }
            }
            if (nodes.length) {
                for (var n = 0, node = nodes[n] ; n < nodes.length; n++, node = nodes[n]) propagate_cause(node);
            } else
                propagate_cause(nodes);
        }
        /**
         * @description Calculate the impact of all of the impacted nodes.
         * @param {GNode|GNode[]} nodes A collection of nodes to update impact. The are json compatible with Cytoscape.js and with NodeProto as prototype.
         */
        function majImpact(nodes) {
            if (nodes.length) majNodesImpact(nodes);
            else              majNodeImpact(nodes);
        }
        /**
         * @description Calculate the impact of a collection of node.
         * @param {GNode[]} nodes
         */
        function majNodesImpact (nodes) {
            for (var n = 0, node = nodes[n]; n < nodes.length; n++, node = nodes[n]) majNodeImpact(node);
        }
        /**
         * @description Calculate the impact of a single node.
         * @param {GNode} node Cytoscape compatible json with NodeProto as prototype.
         */
        function majNodeImpact (node) {
            /**
             * @description Calculate the impact of a node and of it's predecessors recursively.
             * @param {GNode} node_
             */
            function majItsImpact (node_) {
                if (!node_.selected() && node_.areYou(node_p.impacted)) {
                    var impacts = new Impact();
                    impacts.setRedundant(node_.data(node_a.data.redundant));

                    var precedingEdges = node_.precedingEdges();
                    for (var e = 0, precedingEdge = precedingEdges[e]; e < precedingEdges.length ; e++, precedingEdge = precedingEdges[e]) {
                        var preceding_node = precedingEdge.source();
                        /*
                         * Preparing:
                         *       - The source node class to know if this source node has a special treatment;
                         *       - The source node value, 0 if this source node is not impacted, the source node itself otherwise.
                         */
                        var preceding_class = preceding_node.data(node_a.data.class);
                        var preceding_node_value, preceding_edge_value = parseInt(precedingEdge.data(edge_a.data.criticality));
                        if (preceding_node.areYou(node_p.impacted)) {
                            if (!preceding_node.areYou(node_p.selected))            majItsImpact(preceding_node);
                            preceding_node_value = preceding_node.data(node_a.data.impactNode);
                        } else {
                            preceding_node_value = 0;
                        }
                        /* Testing if this source node has special treatment or not. */
                        if (node_.data("nb_" + preceding_class.toLowerCase())) {
                            if (!impacts.specificExist(preceding_class)) {
                                impacts.addSpecific(preceding_class, node_.data("nb_" + preceding_class.toLowerCase()));
                            }
                            impacts.addSpecificValue(preceding_class, preceding_edge_value, preceding_node_value);
                        } else {
                            impacts.addNormalValue(preceding_edge_value, preceding_node_value);
                        }
                    }

                    node_.data(node_a.data.impactNode, impacts.getImpactNodeValue());
                }
            }

            majItsImpact(node);
        }
        /**
         * @description Update the classes of the nodes of the json to use (not the groups).
         */
        function majClasses() {
            var jsonToUse = getJsonToUse();
            var node_iterator = new IteratorGraph(json.origin.nodes);
            node_iterator.forEach(
                /**
                 * @param {GNode} node
                 */
                function (node) {
                    node.eraseClasses();
                    if (node.areYou(node_p.impacted)){
                        node.addClass(css_classes.impacted);
                        node.addClass(getClassColor(node.data(node_a.data.impactNode)));
                    } else if (node.areYou(node_p.causeOf)) {
                        if (node.getProp(node_p.causeOf).length > 1)  node.addClass(css_classes.intersection);
                        else                                          node.addClass(css_classes.cause);
                    } else {
                        node.addClass(css_classes.external);
                    }
            }, function (percentage) { console.log("Update classes: " + percentage.toString().substring(0, 4) + "% clear."); });

            if (jsonToUse.groups.length > 0) {
                node_iterator.changeElement(jsonToUse.groups);
                node_iterator.forEach(
                    /**
                     * @param {GNode} group
                     */
                    function (group) {
                        group.eraseClasses();
                        var nodes = group.getProp(node_p.nodes);
                        var classArr = [],
                            impact = false, cause = false, nothing = false;
                        for (var n = 0, node = nodes[n]; n < nodes.length ; n++, node = nodes[n]) {
                            if (node.areYou(node_p.impacted)) {
                                impact = true;
                                classArr.push(node.data(node_a.data.impactNode));
                            } else if (node.areYou(node_p.causeOf)) {
                                cause = true;
                                if (node.getProp(node_p.causeOf).length > 1) {
                                    if (classArr.indexOf(css_classes.intersection) === -1) classArr.push(css_classes.intersection);
                                } else {
                                    if (classArr.indexOf(css_classes.cause) === -1) classArr.push(css_classes.cause);
                                }
                            } else {
                                nothing = true;
                                if (classArr.indexOf(css_classes.external) === -1) classArr.push(css_classes.external);
                            }
                        }

                        if (impact) {
                            if (cause) {
                                if (classArr.indexOf(css_classes.cause) !== -1) classArr.splice(classArr.indexOf(css_classes.cause), 1);
                                if (classArr.indexOf(css_classes.intersection) !== -1 ) classArr.splice(classArr.indexOf(css_classes.intersection), 1);
                            }
                            if (nothing) classArr.splice(classArr.indexOf(css_classes.external), 1);

                            group.addClass(css_classes.impacted);
                            group.addClass(getClassColor(Math.min.apply(null, classArr)));
                        } else if (cause) {
                            if (classArr.indexOf(css_classes.intersection) !== -1)    group.addClass(css_classes.intersection);
                            else                                                      group.addClass(css_classes.cause);
                        } else if (nothing) {
                            group.addClass(css_classes.external);
                        }
                    }
                );
            }
        }

        initJsonElement();
    }

    return ElementJSON;
}());